"""Datasets available out of the box."""
# pylint: disable=invalid-name,line-too-long

import logging
import re
from os import path
from itertools import product

from glob import glob
import numpy as np
from tqdm import tqdm
import cv2

from .. import utils
from ..core import (Image, Scene, Selection, SceneCollection, Annotation,
                    AnnotationConfiguration)
from .voc import load_voc

log = logging.getLogger(__name__)


def load_voc2012(subset='train') -> SceneCollection:
    """PASCAL VOC 2012

    Args:
        subset: One of `train`, `val`, or `trainval`.
            If `trainval`, the scene collection contains both the
            train and validation sets. If `train`, only the
            training set. If `val` only the validation
            set.

    Returns:
        A scene collection containing the PASCAL VOC 2012
        dataset.
    """
    assert subset in ['train', 'val', 'trainval'], \
        'Subset must be one of train, val, or trainval'

    def extract_check_fn(root_dir):
        image_dir = path.join(root_dir, 'vocdevkit', 'voc2012', 'jpegimages')
        image_files = glob(path.join(image_dir, '*.jpg'))
        annotation_dir = path.join(root_dir, 'vocdevkit', 'voc2012',
                                   'annotations')
        annotation_files = glob(path.join(annotation_dir, '*.xml'))
        return len(annotation_files) == 17125 and len(image_files) == 17125

    root_dir = utils.get_file(
        origin='http://pjreddie.com/media/files/VOCtrainval_11-May-2012.tar',
        file_hash=
        'e14f763270cf193d0b5f74b169f44157a4b0c6efa708f4dd0ff78ee691763bcb',
        cache_subdir=path.join('datasets', 'voc2012'),
        hash_algorithm='sha256',
        extract=True,
        archive_format='tar',
        extract_check_fn=extract_check_fn)
    image_dir = path.join(root_dir, 'vocdevkit', 'voc2012', 'jpegimages')
    annotation_dir = path.join(root_dir, 'vocdevkit', 'voc2012', 'annotations')
    imageset_dir = path.join(root_dir, 'vocdevkit', 'voc2012', 'imagesets',
                             'main')
    sids = []
    if subset in ['trainval', 'train']:
        sid_path = path.join(imageset_dir, 'train.txt')
        with open(sid_path, 'r') as f:
            sids.extend(f.read().split('\n'))
    if subset in ['trainval', 'val']:
        sid_path = path.join(imageset_dir, 'val.txt')
        with open(sid_path, 'r') as f:
            sids.extend(f.read().split('\n'))
    filepaths = [
        path.join(annotation_dir, '{0}.xml'.format(sid)) for sid in sids
        if len(sid.strip()) > 0
    ]
    return load_voc(
        filepaths=filepaths,
        annotation_config=AnnotationConfiguration.VOC,
        image_dir=image_dir)


def load_shapes(
        width=256,
        height=256,
        object_count_bounds=(3, 8),
        object_width_bounds=(20, 40),
        n_scenes=100,
) -> SceneCollection:
    """A simple dataset for testing.

    Args:
        width: The width of each image
        height: The height of each image
        object_count_bounds: A tuple indicating the minimum and maximum
            number of objects in each image
        object_width_bounds: A tuple indicating the minimum and maximum
            widths of the objects in each image
        n_scenes: The number of scenes to generate

    Returns:
        A scene collection of images with circles and rectangles in it.
    """
    annotation_config = AnnotationConfiguration([
        ' '.join([s, c])
        for s, c in product(['RED', 'BLUE', 'GREEN'], ['RECTANGLE', 'CIRCLE'])
    ])

    def make_scene():
        image = Image.new(width=width, height=height, channels=3, cval=255)
        object_count = np.random.randint(*object_count_bounds)
        ws = np.random.randint(*object_width_bounds, size=object_count)
        xs = np.random.randint(
            low=0, high=width - object_width_bounds[-1], size=object_count)
        ys = np.random.randint(
            low=0, high=height - object_width_bounds[-1], size=object_count)
        shapes = np.random.choice(['RECTANGLE', 'CIRCLE'], size=object_count)
        colors = np.random.choice(['RED', 'BLUE', 'GREEN'], size=object_count)
        lookup = {
            'RED': (255, 0, 0),
            'BLUE': (0, 0, 255),
            'GREEN': (0, 255, 0)
        }
        annotations = []
        for x, y, w, shape, color in zip(xs, ys, ws, shapes, colors):
            if shape == 'RECTANGLE':
                cv2.rectangle(  # pylint: disable=no-member
                    image,
                    pt1=(x, y),
                    pt2=(x + w, y + w),
                    thickness=-1,
                    color=lookup[color])
            elif shape == 'CIRCLE':
                r = w // 2
                w = 2 * r
                cv2.circle(  # pylint: disable=no-member
                    image,
                    center=(x + r, y + r),
                    radius=r,
                    thickness=-1,
                    color=lookup[color])
            annotations.append(
                Annotation(
                    selection=Selection([[x, y], [x + w, y + w]]),
                    category=annotation_config[' '.join([color, shape])]))
        return Scene(
            annotations=annotations,
            image=image,
            annotation_config=annotation_config)

    scenes = [make_scene() for n in range(n_scenes)]
    return SceneCollection(scenes=scenes, annotation_config=annotation_config)


def load_oxfordiiitpets(breed=True) -> SceneCollection:
    """Load the Oxford-IIIT pets dataset. It is not divided into
    train, validation, and test because it appeared some files were missing
    from the trainval and test set documents
    (e.g., english_cocker_spaniel_164).

    Args:
        breed: Whether to use the breeds as the class labels. If False, the
            class labels are limited to dog or cat.

    Returns:
        A scene collection containing the dataset
    """
    image_dir = utils.get_file(
        origin='http://www.robots.ox.ac.uk/~vgg/data/pets/data/images.tar.gz',
        file_hash=
        '67195c5e1c01f1ab5f9b6a5d22b8c27a580d896ece458917e61d459337fa318d',
        cache_subdir=path.join('datasets', 'oxfordiiitpets'),
        hash_algorithm='sha256',
        extract=True,
        archive_format='tar',
        extract_check_fn=
        lambda directory: len(glob(path.join(directory, 'images', '*.jpg'))) == 7390
    )
    annotations_dir = utils.get_file(
        origin='http://www.robots.ox.ac.uk/~vgg/data/pets/data/annotations.tar.gz',
        file_hash='52425fb6de5c424942b7626b428656fcbd798db970a937df61750c0f1d358e91',
        cache_subdir=path.join('datasets', 'oxfordiiitpets'),
        hash_algorithm='sha256',
        extract=True,
        archive_format='tar',
        extract_check_fn=lambda directory: len(glob(path.join(directory, 'annotations', 'xmls', '*.xml'))) == 3686
    )
    filepaths = glob(
        path.join(annotations_dir, 'annotations', 'xmls', '*.xml'))
    image_dir = path.join(image_dir, 'images')
    collection = load_voc(
        filepaths=filepaths,
        annotation_config=AnnotationConfiguration(['dog', 'cat']),
        image_dir=image_dir)
    if not breed:
        return collection
    assert all(len(s.annotations) in [1, 2] for s in collection.scenes), \
        'An error occurred handling pets dataset'
    labels = [
        '_'.join(path.splitext(path.split(f)[1])[0].split('_')[:-1])
        for f in filepaths
    ]
    annotation_config = AnnotationConfiguration(sorted(set(labels)))
    return SceneCollection(
        scenes=[
            scene.assign(
                annotations=[
                    a.assign(category=annotation_config[label])
                    for a in scene.annotations
                ],
                annotation_config=annotation_config)
            for scene, label in zip(collection.scenes, labels)
        ],
        annotation_config=annotation_config)


def load_icdar2015(subset: str = 'train',
                   text_category: str = 'text') -> SceneCollection:
    """Loads dataset from 2015 Robust Reading Competition.
    More details available at http://rrc.cvc.uab.es/?ch=4&com=introduction

    Args:
        subset: One of `train`, `test`, or `traintest`.
            If `traintest`, the scene collection contains both the
            train and test sets. If `train`, only the
            training set. If `test` only the test set.
        text_category: The category name to use for the
            annotation configuration.

    Returns:
        A scene collection containing the ICDAR 2015 dataset
        dataset.
    """

    scenes = []
    annotation_config = AnnotationConfiguration([text_category])

    def annotation_from_box(box):
        # We have to do this because there are no quotes
        # around text that contains commas in the annotation files.
        # For an example, see gt_img_115.txt.
        elements = box.split(',')
        x1, y1, x2, y2, x3, y3, x4, y4 = map(int, elements[:8])
        text = ','.join(elements[8:])  # pylint: disable=unused-variable
        return Annotation(
            selection=Selection(
                np.array([[x1, y1], [x2, y2], [x3, y3], [x4, y4]])),
            category=annotation_config[text_category])

    def scene_from_files(annotation_file, image_file):
        with open(annotation_file, 'r', encoding='utf-8-sig') as f:
            annotations = [
                annotation_from_box(box) for box in f.read().split('\n')[:-1]
            ]
        return Scene(
            image=image_file,
            annotations=annotations,
            annotation_config=annotation_config)

    def scenes_from_directories(annotations_dir, images_dir):
        pattern = re.compile(r'img_([0-9]+).')
        image_files = sorted(
            glob(path.join(images_dir, '*.jpg')),
            key=lambda x: pattern.findall(x)[0])
        annotation_files = sorted(
            glob(path.join(annotations_dir, '*.txt')),
            key=lambda x: pattern.findall(x)[0])
        assert len(image_files) == len(annotation_files), \
            'An error occurred loading the dataset.'
        scenes = []
        for annotation_file, image_file in tqdm(
                zip(annotation_files, image_files),
                desc='Creating scenes from ' + annotations_dir,
                total=len(annotation_files)):
            scenes.append(scene_from_files(annotation_file, image_file))
        return scenes

    if subset in ['train', 'traintest']:
        images_dir = utils.get_file(
            origin=
            'https://storage.googleapis.com/miradata/datasets/rrc2015/ch4_training_images.zip',
            file_hash=
            '07e2d816dab67df4cd509641b1e1e8c720d3b15873a65b6dc9c1aa6a4de3d2c6',
            cache_subdir=path.join('datasets', 'rrc2015'),
            hash_algorithm='sha256',
            extract=True,
            archive_format='zip',
            extract_check_fn=
            lambda directory: len(glob(path.join(directory, '*.jpg'))) == 1000)
        annotations_dir = utils.get_file(
            origin=
            'https://storage.googleapis.com/miradata/datasets/rrc2015/ch4_training_localization_transcription_gt.zip',
            file_hash=
            'b9f2f0343d016a326bcafe3c28e0ddbda18b3a1e4a8a595784ba9dc3e305d754',
            cache_subdir=path.join('datasets', 'rrc2015'),
            hash_algorithm='sha256',
            extract=True,
            archive_format='zip',
            extract_check_fn=
            lambda directory: len(glob(path.join(directory, '*.txt'))) == 1000)
        scenes += scenes_from_directories(annotations_dir, images_dir)
    if subset in ['test', 'traintest']:
        images_dir = utils.get_file(
            origin=
            'https://storage.googleapis.com/miradata/datasets/rrc2015/ch4_test_images.zip',
            file_hash=
            'ecfb2488333372b7381b14ccb5ac2127de88fb935f6759fe89e5c041b0e87358',
            cache_subdir=path.join('datasets', 'rrc2015'),
            hash_algorithm='sha256',
            extract=True,
            archive_format='zip',
            extract_check_fn=
            lambda directory: len(glob(path.join(directory, '*.jpg'))) == 500)
        annotations_dir = utils.get_file(
            origin=
            'https://storage.googleapis.com/miradata/datasets/rrc2015/Challenge4_Test_Task1_GT.zip',
            file_hash=
            '8f434c410cfcf680b8421e3c22abf31b9297d8e255c3b13acc2890fddc6db6b0',
            cache_subdir=path.join('datasets', 'rrc2015'),
            hash_algorithm='sha256',
            extract=True,
            archive_format='zip',
            extract_check_fn=
            lambda directory: len(glob(path.join(directory, '*.txt'))) == 500)
        scenes += scenes_from_directories(annotations_dir, images_dir)
    return SceneCollection(annotation_config=annotation_config, scenes=scenes)



"""VOC parsing tools"""

import os
from typing import List
from xml.etree import ElementTree
import logging

from tqdm import tqdm

from ..core import (Selection, Scene, SceneCollection, AnnotationConfiguration,
                    Annotation)

log = logging.getLogger(__name__)  # pylint: disable=invalid-name

VOC_SCENE_METADATA_MAP = [['folder'], ['filename'], ['size', 'width'],
                          ['size', 'height'], ['source', 'database'],
                          ['source', 'annotation'], ['source', 'image'],
                          ['segmented']]

VOC_ANNOTATION_METADATA_MAP = [['pose'], ['truncated'], ['difficult']]


def map_xml_to_metadata(paths: List[List[str]], root: ElementTree.Element):
    metadata = {}
    for path in paths:
        key = ':'.join(path)
        elem = root
        for part in path:
            elem = elem.find(part)
            if elem is None:
                log.info('Missing annotation metadata: %s', ':'.join(path))
                break
        if elem is None:
            continue
        metadata[key] = elem.text
    return metadata


def load_voc(
        filepaths: str,
        annotation_config: AnnotationConfiguration,
        image_dir: str = None,
) -> SceneCollection:
    """Read a scene from a VOC XML annotation file. Remaining arguments
    passed to scene constructor.

    Args:
        filepaths: A list of VOC files to read
        image_folder: Folder in which to look for images. Defaults to same
            folder as XML file prepended to the folder specified in the
            XML file.
        annotation_config: The annotation configuration to use.

    Returns:
        A new scene collection, one scene per VOC file
    """
    scenes = []
    for filepath in tqdm(filepaths, desc='Loading VOC annotation files.'):
        annotations = []
        root = ElementTree.parse(filepath).getroot()

        # Get the scene level metadata
        scene_metadata = map_xml_to_metadata(
            paths=VOC_SCENE_METADATA_MAP, root=root)

        if image_dir is None:
            folder = scene_metadata['folder']
            if folder is None:
                image_dir = os.path.dirname(filepath)
            else:
                image_dir = os.path.join(os.path.dirname(filepath), folder)

        image_path = os.path.join(image_dir, scene_metadata['filename'])
        for obj in root.findall('object'):
            category = annotation_config[obj.find('name').text]
            selection = None
            for bndbox in obj.findall('bndbox'):
                xmin, ymin, xmax, ymax = [
                    int(float(bndbox.find(k).text))
                    for k in ['xmin', 'ymin', 'xmax', 'ymax']
                ]
                current = Selection([[xmin, ymin], [xmax, ymax]])
                if selection is None:
                    selection = current
                else:
                    selection += current
            annotations.append(
                Annotation(selection=selection, category=category))
        scenes.append(
            Scene(
                annotation_config=annotation_config,
                annotations=annotations,
                image=image_path))
    return SceneCollection(scenes=scenes, annotation_config=annotation_config)



"""COCO parsing tools"""
# pylint: disable=invalid-name

import logging
from os import path
import json

from tqdm import tqdm
import numpy as np

from ..core import (Scene, SceneCollection, AnnotationConfiguration,
                    Annotation, Selection)

log = logging.getLogger(__name__)


def load_coco(
        annotations_file: str,
        image_dir: str,
        annotation_config: AnnotationConfiguration = None,
) -> SceneCollection:
    """Obtain a scene collection from a COCO JSON file.

    Args:
        annotations_file: The annotation file to load
        image_dir: The directory in which to look for images
        annotation_config: The annotation configuration to
            use. If None, it is inferred from the annotations
            category file.

    Returns:
        A scene collection
    """
    if image_dir is None:
        image_dir = path.split(annotations_file)[0]
    with open(annotations_file, 'r') as f:
        data = json.load(f)
    categories = {}
    for category in data['categories']:
        categories[category['id']] = category['name']
    category_names = [
        c[1] for c in sorted(list(categories.items()), key=lambda x: x[0])
    ]
    if annotation_config is None:
        annotation_config = AnnotationConfiguration(category_names)
    assert len(categories) == len(annotation_config), \
        'Annotation configuration incompatible with in-file categories'
    assert all([c in annotation_config for c in category_names]), \
        'Some in-file categories not in annotation configuration'
    assert all([c.name in category_names for c in annotation_config]), \
        'Some annotation configuration categories not in file'

    annotations = np.array([[ann['image_id'], ann['category_id']] + ann['bbox']
                            for ann in data['annotations']])
    annotations = annotations[annotations[:, 0].argsort()]
    images = sorted(data['images'], key=lambda x: x['id'])
    del data
    scenes = [None] * len(images)
    startIdx = 0
    for imageIdx, image in tqdm(
            enumerate(images), total=len(images), desc='Creating scenes'):
        current = annotations[startIdx:][annotations[startIdx:, 0] ==
                                         image['id'], 1:]
        startIdx += len(current)

        scenes[imageIdx] = Scene(
            image=path.join(image_dir, image['file_name']),
            annotation_config=annotation_config,
            annotations=[
                Annotation(
                    category=annotation_config[categories[int(ann[0])]],
                    selection=Selection(
                        points=[[ann[1], ann[2]],
                                [ann[1] + ann[3], ann[2] + ann[4]]]))
                for ann in current
            ])
    return SceneCollection(scenes=scenes, annotation_config=annotation_config)


def load_coco_text(
        annotations_file: str,
        image_dir: str,
        annotation_config: AnnotationConfiguration = None,
) -> SceneCollection:
    """Obtain a scene collection from a COCO Text JSON file
    (e.g., that which can be obtained from https://bgshih.github.io/cocotext/)

    Args:
        annotations_file: The annotation file to load
        image_dir: The directory in which to look for images
        annotation_config: The annotation configuration to
            use. If None, it is inferred from the annotations
            category file.

    Returns:
        A scene collection
    """
    with open(annotations_file, 'r') as f:
        data = json.load(f)

    category_names = set([ann['class'] for ann in data['anns'].values()])
    if annotation_config is None:
        annotation_config = AnnotationConfiguration(
            sorted(list(category_names)))  # noqa: E501
    assert len(category_names) == len(annotation_config), \
        'Annotation configuration incompatible with in-file categories'
    assert all([c in annotation_config for c in category_names]), \
        'Some in-file categories not in annotation configuration'
    assert all([c.name in category_names for c in annotation_config]), \
        'Some annotation configuration categories not in file'

    images = data['imgs']
    scenes = [None] * len(images)

    for sceneIdx, (imageId, imageData) in tqdm(
            enumerate(images.items()), total=len(images),
            desc='Creating scenes'):
        anns = [
            data['anns'][str(annId)] for annId in data['imgToAnns'][imageId]
        ]  # noqa: E501
        scenes[sceneIdx] = Scene(
            annotation_config=annotation_config,
            image=path.join(image_dir, imageData['file_name']),
            annotations=[
                Annotation(
                    category=annotation_config[ann['class']],
                    selection=Selection([
                        [ann['bbox'][0], ann['bbox'][1]],
                        [
                            ann['bbox'][0] + ann['bbox'][2],
                            ann['bbox'][1] + ann['bbox'][3]
                        ]  # noqa: E501
                    ])) for ann in anns
            ])
    return SceneCollection(scenes=scenes, annotation_config=annotation_config)



"""Core utilities"""

import logging
import os
import hashlib
import io
import shutil
import tarfile
import zipfile
import urllib

import six
import numpy as np
from tqdm import tqdm

try:
    from keras_retinanet.utils.compute_overlap import compute_overlap  # noqa: F401, E501
except ValueError:
    # Avoids crashing documentation due to numpy
    # warning.
    def compute_overlap():
        pass


log = logging.getLogger(__name__)

# Progbar, _extract_archive, get_file, _hash_file and validate_file
# adapted / copied from Keras to avoid requiring
# TensorFlow to simply use file downloads, etc.

# ProgressFileObject adapted from https://stackoverflow.com/a/3668977


def get_datadir_base():
    cache_dir_env = os.environ.get('MIRA_CACHE')
    if cache_dir_env is not None:
        cache_dir = os.path.abspath(cache_dir_env)
    else:
        cache_dir = os.path.join(os.path.expanduser('~'), '.mira')
    return os.path.expanduser(cache_dir)


class ProgressFileObject(io.FileIO):  # noqa: E302
    def __init__(self, path, *args, **kwargs):
        self._total_size = os.path.getsize(path)
        self.progbar = tqdm(total=self._total_size, desc='Reading ' + path)
        io.FileIO.__init__(self, path, *args, **kwargs)

    def read(self, size=-1):
        self.progbar.update(size)
        return io.FileIO.read(self, size)

    def close(self):
        self.progbar.close()
        super().close()


def _extract_archive(  # noqa: E302
        file_path,
        path='.',
        archive_format='auto',
        extract_check_fn=None):
    """Extracts an archive if it matches tar, tar.gz, tar.bz, or zip formats.

    # Arguments
        file_path: path to the archive file
        path: path to extract the archive file
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.

    # Returns
        True if a match was found and an archive extraction was completed,
        False otherwise.
    """
    if archive_format is None:
        return False
    if archive_format == 'auto':
        archive_format = ['tar', 'zip']
    if isinstance(archive_format, six.string_types):
        archive_format = [archive_format]

    for archive_type in archive_format:
        if archive_type == 'tar':
            # Implement progress bar for tar files,
            # per https://stackoverflow.com/questions/3667865/python-tarfile-progress-output  # pylint: disable=line-too-long
            # open_fn = tarfile.open
            open_fn = lambda fp: tarfile.open(fileobj=ProgressFileObject(fp))  # noqa: E731,E501
            is_match_fn = tarfile.is_tarfile
        if archive_type == 'zip':
            # open_fn = zipfile.ZipFile
            open_fn = lambda fp: zipfile.ZipFile(file=ProgressFileObject(fp))  # noqa: E731,E501
            is_match_fn = zipfile.is_zipfile

        if is_match_fn(file_path):
            if (os.path.exists(path) and extract_check_fn is not None):
                log.info('Found target folder. Checking for '
                         'existing complete extraction.')
                if extract_check_fn(path):
                    log.info('Found complete extraction.')
                    return True
                else:
                    log.info('Extraction is not complete. Re-extracting.')
            log.info('Extracting {0} to {1}'.format(file_path,
                                                    path))  # noqa: E501
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(path)
                except (tarfile.TarError, RuntimeError, KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False


def get_file(origin,
             fname=None,
             file_hash=None,
             cache_subdir='datasets',
             hash_algorithm='auto',
             extract=False,
             archive_format='auto',
             extract_check_fn=None):
    """Downloads a file from a URL if it not already in the cache.

    By default the file at the url `origin` is downloaded to the
    cache_dir `~/.keras`, placed in the cache_subdir `datasets`,
    and given the filename `fname`. The final location of a file
    `example.txt` would therefore be `~/.keras/datasets/example.txt`.

    Files in tar, tar.gz, tar.bz, and zip formats can also be extracted.
    Passing a hash will verify the file after download. The command line
    programs `shasum` and `sha256sum` can compute the hash.

    # Arguments
        fname: Name of the file. If an absolute path `/path/to/file.txt` is
            specified the file will be saved at that location. If None,
            the name of the file at origin is used instead.
        origin: Original URL of the file.
        untar: Deprecated in favor of 'extract'.
            boolean, whether the file should be decompressed
        file_hash: The expected hash string of the file after download.
            The sha256 and md5 hash algorithms are both supported.
        cache_subdir: Subdirectory under the Keras cache dir where the file is
            saved. If an absolute path `/path/to/folder` is
            specified the file will be saved at that location.
        hash_algorithm: Select the hash algorithm to verify the file.
            options are 'md5', 'sha256', and 'auto'.
            The default 'auto' detects the hash algorithm in use.
        extract: True tries extracting the file as an Archive, like tar or zip.
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.

    # Returns
        Path to the downloaded file
    """
    if fname is None:
        fname = os.path.basename(urllib.parse.urlparse(origin).path)
    datadir_base = get_datadir_base()
    datadir = os.path.join(datadir_base, cache_subdir)
    if not os.path.exists(datadir):
        os.makedirs(datadir)

    fpath = os.path.join(datadir, fname)

    download = False
    if os.path.exists(fpath):
        # File found; verify integrity if a hash was provided.
        if file_hash is not None:
            if not validate_file(fpath, file_hash, algorithm=hash_algorithm):
                log.warning(
                    'A local file was found, but it seems to be '
                    'incomplete or outdated because the ' + hash_algorithm +
                    ' file hash does not match the original value of ' +
                    file_hash + ' so we will re-download the data.')
                download = True
    else:
        download = True

    if download:

        class ProgressTracker(object):
            # Maintain progbar for the lifetime of download.
            # This design was chosen for Python 2.7 compatibility.
            progbar = None

        def dl_progress(count, block_size, total_size):
            if ProgressTracker.progbar is None:
                if total_size == -1:
                    total_size = None
                ProgressTracker.progbar = tqdm(
                    total=total_size, desc='Downloading data from ' + origin)
            else:
                ProgressTracker.progbar.update(block_size)

        error_msg = 'URL fetch failure on {} : {} -- {}'
        try:
            try:
                urllib.request.urlretrieve(origin, fpath, dl_progress)
            except urllib.error.HTTPError as e:  # pylint: disable=invalid-name
                raise Exception(error_msg.format(origin, e.code, e.msg))
            except urllib.error.URLError as e:  # pylint: disable=invalid-name
                raise Exception(error_msg.format(origin, e.errno, e.reason))
        except (Exception, KeyboardInterrupt):
            if os.path.exists(fpath):
                os.remove(fpath)
            raise
        ProgressTracker.progbar.close()
        ProgressTracker.progbar = None

    if extract:
        extract_target = os.path.join(
            datadir,
            os.path.split(fpath)[1].split(os.extsep)[0])
        extract_result = _extract_archive(
            file_path=fpath,
            path=extract_target,
            archive_format=archive_format,
            extract_check_fn=extract_check_fn)
        if extract_result:
            fpath = extract_target

    return fpath


def _hash_file(fpath, algorithm='sha256', chunk_size=65535):
    """Calculates a file sha256 or md5 hash.

    # Example

    ```python
        >>> from keras.data_utils import _hash_file
        >>> _hash_file('/path/to/file.zip')
        'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    ```

    # Arguments
        fpath: path to the file being validated
        algorithm: hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    # Returns
        The file hash
    """
    if (algorithm == 'sha256') or (algorithm == 'auto' and len(hash) == 64):
        hasher = hashlib.sha256()
    else:
        hasher = hashlib.md5()

    with open(fpath, 'rb') as fpath_file:
        for chunk in iter(lambda: fpath_file.read(chunk_size), b''):
            hasher.update(chunk)

    return hasher.hexdigest()


def validate_file(fpath, file_hash, algorithm='auto', chunk_size=65535):
    """Validates a file against a sha256 or md5 hash.

    # Arguments
        fpath: path to the file being validated
        file_hash:  The expected hash string of the file.
            The sha256 and md5 hash algorithms are both supported.
        algorithm: Hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    # Returns
        Whether the file is valid
    """
    if ((algorithm == 'sha256')
            or (algorithm == 'auto' and len(file_hash) == 64)):
        hasher = 'sha256'
    else:
        hasher = 'md5'
    log.info('Checking hash for {0}.'.format(os.path.split(fpath)[1]))
    if str(_hash_file(fpath, hasher, chunk_size)) == str(file_hash):
        return True
    else:
        return False



"""
Tools for loading and saving VGG Image Annotator datasets.
"""

import os
import json
import uuid
import typing
import logging

from tqdm import tqdm
import pandas as pd

from .. import core

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def load_via(project_file: str,
             label_key: str = 'class',
             image_dir: str = None,
             cache: typing.Union[str, bool] = True) -> core.SceneCollection:
    """Load annotations created by the `VGG Image Annotation Tool
    <http://www.robots.ox.ac.uk/~vgg/software/via/>`_.

    Args:
        project_file: The file containing the
            annotations. Only the JSON
            format is supported.
        label_key: The attribute name to use for class labels
        image_dir: The directory in which to look
            for images. If None, defaults the
            folder in which the annotations_file
            is stored.
        cache: The cache parameter to pass to `Scene`. Useful
            if the annotations file references URLs (i.e.,
            specify a directory in which to save images).
    """
    if image_dir is None:
        image_dir = os.path.split(project_file)[0]
    if not project_file.endswith('.json'):
        raise NotImplementedError('Only the VIA JSON format is supported.')
    with open(project_file) as f:  # pylint: disable=invalid-name
        project_data = json.loads(f.read())
    img_metadata = project_data['_via_img_metadata'].values()
    project_attrs = project_data['_via_attributes']
    regions_df = pd.concat([
        pd.io.json.json_normalize(img['regions']).assign(
            filename=img['filename']
        ) for img in img_metadata
    ], axis=0, sort=False)
    # We may not have any regions!
    if len(regions_df.index) > 0:
        bad_filenames = regions_df[
            regions_df[f'region_attributes.{label_key}'].isnull()
        ]['filename'].unique()
    else:
        bad_filenames = []
    if len(bad_filenames) > 0:  # pylint: disable=len-as-condition
        log.warning(
            'The following files, which will be skipped, have missing labels: %s',
            ', '.join(bad_filenames)
        )
    filenames_metadata = [
        (img['filename'], {**img['file_attributes'], **{'filename': img['filename']}})
        for img in img_metadata if img['filename'] not in bad_filenames
    ]
    regions_df = regions_df[~regions_df.filename.isin(bad_filenames)]

    annotation_config = core.AnnotationConfiguration(
        project_attrs['region'][label_key]['options'].keys()
    )
    assert len(regions_df) == 0 or (regions_df['shape_attributes.name'] == 'rect').all(), \
        'Only axis aligned rectangular selections are supported.'
    return core.SceneCollection(
        scenes=[
            core.Scene(
                image=os.path.join(image_dir, filename),
                annotation_config=annotation_config,
                cache=cache,
                metadata=metadata,
                annotations=[
                    core.Annotation(
                        selection=core.Selection(
                            points=[
                                [r['shape_attributes.x'], r['shape_attributes.y']],
                                [
                                    r['shape_attributes.x'] + r['shape_attributes.width'],
                                    r['shape_attributes.y'] + r['shape_attributes.height']
                                ]
                            ]
                        ),
                        category=annotation_config[r[f'region_attributes.{label_key}']]
                    )
                    for _, r in regions_df[regions_df.filename == filename].iterrows()
                ]
            ) for filename, metadata in filenames_metadata
        ],
        annotation_config=annotation_config
    )

def save_via(
        collection: core.SceneCollection,
        export_dir: str,
        label_key='class',
        project_name='mira_export',
        filenames: typing.List[str] = None
):
    """Save a scene collection in VIA format.

    Args:
        collection: The scene collection to save.
        export_dir: The directory where the via.json file along with all images
            will be saved.
        label_key: The region attribute in which to store the labels
    """
    # Make sure the directory does not exist.
    if os.path.isdir(export_dir):
        raise ValueError(f'{export_dir} already exists.')
    os.makedirs(export_dir, exist_ok=False)
    if filenames is not None:
        assert len(filenames) == len(collection), \
            'Length of filenames must be equal to length of collection'
    else:
        filenames = [scene.metadata.get('filename', f'{uuid.uuid4()}.jpg') for scene in collection]
    assert len(set(filenames)) == len(filenames), 'Filenames must be unique.'
    data = [(filename, scene) for scene, filename in zip(collection, filenames)]
    metadata_keys = []
    for filename, scene in tqdm(data, desc='Saving images'):
        scene.image.save(os.path.join(export_dir, filename))
        if scene.metadata is not None:
            metadata_keys.extend(list(scene.metadata.keys()))
    metadata_keys = list(set(metadata_keys))
    data = [
        (filename, scene, os.path.getsize(os.path.join(export_dir, filename)))
        for filename, scene in data
    ]
    sc_bboxes = [[ann.selection.xywh() for ann in s.annotations] for s in collection]
    img_metadata_keys = [f'{filename}{size}' for filename, _, size in data]
    img_metadata_values = [
        {
            'filename': filename,
            'size': size,
            "file_attributes": {} if scene.metadata is None else dict(
                zip(
                    metadata_keys, [scene.metadata.get(k, '') for k in metadata_keys]
                )
            ),
            'regions': [
                {
                    'shape_attributes': {
                        'name': 'rect',
                        'x': x, 'y': y, 'width': width, 'height': height
                    },
                    'region_attributes': {
                        label_key: ann.category.name
                    }
                } for (x, y, width, height), ann in zip(bboxes, scene.annotations)
            ]
        } for (filename, scene, size), bboxes in zip(data, sc_bboxes)
    ]
    via_attributes = {
        'region': {
            label_key: {
                'type': 'dropdown',
                'description': '',
                'options': dict(
                    zip(
                        [category.name for category in collection.annotation_config],
                        ["" for category in collection.annotation_config]
                    )
                ),
                'default_options': {}
            }
        },
        'file': dict(
            zip(
                metadata_keys,
                [
                    {
                        'type': 'text',
                        'description': '',
                        'default_value': ''
                    } for k in metadata_keys
                ]
            )
        ) if metadata_keys else {}
    }
    export_object = {
        '_via_img_metadata': dict(zip(img_metadata_keys, img_metadata_values)),
        '_via_attributes': via_attributes,
        "_via_settings": {
            "ui": {
                "annotation_editor_height": 25,
                "annotation_editor_fontsize": 0.8,
                "leftsidebar_width": 18,
                "image_grid": {
                    "img_height": 80,
                    "rshape_fill": "none",
                    "rshape_fill_opacity": 0.3,
                    "rshape_stroke": "yellow",
                    "rshape_stroke_width": 2,
                    "show_region_shape": True,
                    "show_image_policy": "all"
                },
                "image": {
                    "region_label": "__via_region_id__",
                    "region_color": "__via_default_region_color__",
                    "region_label_font": "10px Sans",
                    "on_image_annotation_editor_placement": "NEAR_REGION"
                }
            },
            "core": {
                "buffer_size": 18,
                "filepath": {},
                "default_filepath": ""
            },
            "project": {
                "name": project_name
            }
        }
    }
    with open(os.path.join(export_dir, 'via.json'), 'w') as f:  # pylint: disable=invalid-name
        f.write(json.dumps(export_object))



"""Test mAP calculation."""
# pylint: disable=invalid-name

import numpy as np

from mira import core, metrics


def test_mAP():
    """Make sure mAP makes sense."""
    image = np.random.randint(
        low=0, high=255, size=(100, 100, 3)).astype('uint8')

    annotation_config = core.AnnotationConfiguration(['foo', 'bar', 'baz'])
    true = core.Scene(
        image=image,
        annotation_config=annotation_config,
        annotations=[
            core.Annotation(
                selection=core.Selection([[0, 0], [10, 10]]),
                category=annotation_config['foo']),
            core.Annotation(
                selection=core.Selection([[40, 40], [50, 50]]),
                category=annotation_config['baz']),
        ])
    pred = core.Scene(
        image=image,
        annotation_config=annotation_config,
        annotations=[
            core.Annotation(
                selection=core.Selection([[0, 0], [10, 10]]),
                category=annotation_config['foo'],
                score=0.9),
            core.Annotation(
                selection=core.Selection([[15, 15], [30, 30]]),
                category=annotation_config['foo'],
                score=0.5),
            core.Annotation(
                selection=core.Selection([[45, 45], [50, 50]]),
                category=annotation_config['baz'],
                score=0.8)
        ])

    true_collection = core.SceneCollection([true],
                                           annotation_config=annotation_config)
    pred_collection = core.SceneCollection([pred],
                                           annotation_config=annotation_config)

    maps1 = metrics.mAP(true_collection, pred_collection, iou_threshold=0.25)
    maps2 = metrics.mAP(true_collection, pred_collection, iou_threshold=0.3)

    assert maps1['foo'] == 1
    assert maps2['foo'] == 1
    assert ~np.isfinite(maps1['bar'])
    assert ~np.isfinite(maps2['bar'])
    assert maps1['baz'] == 1  # The IoU is just enough
    assert maps2['baz'] == 0  # The IoU is too small



# pylint: disable=invalid-name
"""Metrics for object detection tasks."""
from typing import List, Tuple

import numpy as np

from ..utils import compute_overlap
from ..core import SceneCollection


def precision_recall_curve(true_collection: SceneCollection,
                           pred_collection: SceneCollection,
                           iou_threshold: float = 0.5) -> List[np.ndarray]:
    """Compute the precision-recall curve for each of the
    classes.

    Args:
        true_collection: The true scene collection
        pred_collection: The predicted scene collection
        iou_threshold: The threshold for detection

    Returns:
        A dict with category names as keys and array of shape (Ni, 3)
        which is the precision, recall, and score for each of the
        predicted boxes for the category.
    """
    assert true_collection.annotation_config == pred_collection.annotation_config, (  # noqa: E501
        'Annotation configurations must match')
    annotation_config = true_collection.annotation_config
    assert len(true_collection.scenes) == len(pred_collection.scenes), \
        'Must have same scenes in each collection'

    # The ith entry in tfs is a list of lists, each of length three,
    # which are the change in the number of true positives and
    # false positives, along with the score at which the change
    # occurred for the ith class.
    tfs = [[[], [], []] for c in range(len(annotation_config))]

    # The ith entry in tfs is the number of true boxes
    # for the ith class.
    pos = [0 for c in range(len(annotation_config))]

    for true, pred in zip(true_collection, pred_collection):
        pred_bboxes = pred.bboxes()
        true_bboxes = true.bboxes()
        pred_scores = pred.scores()
        assert all(s is not None
                   for s in pred_scores), 'All annotations must have a score.'

        for classIdx in range(len(annotation_config)):
            pred_bboxes_cur = pred_bboxes[pred_bboxes[:, 4] == classIdx]
            true_bboxes_cur = true_bboxes[true_bboxes[:, 4] == classIdx]
            pred_scores_cur = pred_scores[pred_bboxes[:, 4] == classIdx]

            nPredicted = len(pred_bboxes_cur)
            nTrue = len(true_bboxes_cur)
            pos[classIdx] += nTrue

            if nPredicted == 0:
                # We have no new information to add if there were no
                # predicted boxes.
                continue

            if nTrue == 0:
                # All of them are false positives
                for score in pred_scores_cur:
                    tfs[classIdx][0].append(0)
                    tfs[classIdx][1].append(1)
                    tfs[classIdx][2].append(score)
                continue

            # Sort the predicted boxes by decreasing confidence
            pred_bboxes_cur = pred_bboxes_cur[(-pred_scores_cur).argsort()]
            pred_scores_cur = pred_scores_cur[(-pred_scores_cur).argsort()]

            # (n, m): status for ith prediction for jth true box
            det = compute_overlap(
                boxes=pred_bboxes_cur.astype('float64'),
                query_boxes=true_bboxes_cur.astype('float64')) > iou_threshold

            fp_prev = 0
            tp_prev = 0
            for i in range(nPredicted):
                tp_cur = det[:i + 1].max(axis=0).sum()
                fp_cur = (i + 1) - det[:i + 1].max(axis=1).sum()

                tp_delta = tp_cur - tp_prev
                fp_delta = fp_cur - fp_prev

                assert tp_delta >= 0
                assert fp_delta >= 0
                assert tp_cur <= nTrue
                assert fp_cur <= nPredicted

                tp_prev = tp_cur
                fp_prev = fp_cur

                tfs[classIdx][0].append(tp_delta)
                tfs[classIdx][1].append(fp_delta)
                tfs[classIdx][2].append(pred_scores_cur[i])

    prs = [None for n in range(len(annotation_config))]

    for classIdx, tfs_cur, pos_cur in zip(
            range(len(annotation_config)), tfs, pos):
        # If we had no detections AND there
        # were no true boxes, precision and recall
        # are not defined.
        tfs_cur = np.array(tfs_cur).T
        tfs_cur = tfs_cur[(-tfs_cur[:, 2]).argsort()]
        tp = tfs_cur[:, 0].cumsum()
        fp = tfs_cur[:, 1].cumsum()
        scores = tfs_cur[:, 2]

        precisions = tp / (tp + fp)
        recalls = tp / pos_cur
        prs[classIdx] = np.vstack([precisions, recalls, scores]).T
    return dict(zip([c.name for c in annotation_config], prs))


def mAP(true_collection: SceneCollection,
        pred_collection: SceneCollection,
        iou_threshold: float = 0.5) -> float:
    """Compute mAP (mean average precision) for
    a pair of scene collections.

    Args:
        true_collection: The true scene collection
        pred_collection: The predicted scene collection
        iou_threshold: The threshold for detection

    Returns:
        mAP class scores
    """
    prs = precision_recall_curve(true_collection, pred_collection,
                                 iou_threshold)
    aps = {}
    for className, prs_cur in prs.items():
        ps = prs_cur[:, 0]
        rs = prs_cur[:, 1]
        pi = np.zeros(11)
        # If rs is None, there were no detections and no true
        # boxes. If it is all nans, then there were no
        # true boxes but there were detections.
        if rs is None or np.isnan(rs).sum() == rs.shape[0]:
            aps[className] = np.nan
            continue
        for i, r in enumerate(np.linspace(0, 1, 11)):
            # From section 4.2 of VOC paper,the precision at each
            # recall level r is interpolated by taking the maximum
            # precision measured for a method for which
            # the corresponding recall exceeds r
            pc = ps[rs >= r]
            if len(pc) > 0:  # pylint: disable=len-as-condition
                pi[i] = pc.max()
        aps[className] = pi.mean()
    return aps
